TargetWindow = class(Turbine.UI.Window)
function TargetWindow:Constructor()
    Turbine.UI.Window.Constructor(self)
    self.width = Settings.TWidth
    self.nameHeight = 20
    self.dataHeight = 20
    self.borderWidth = 1
    self.height = self.nameHeight + self.dataHeight + 2 * self.borderWidth
    self.x = Settings.TX
    self.y = Settings.TY

    self.Timers = {}

    if Settings.TargetWindowToken < 1 or Settings.NewTargetWindowEffects == nil then
        Settings.NewTargetWindowEffects = SetEffects()
        Settings.TargetWindowToken = 1
    end

    self.UpdateDelay = 0.05
    self.CheckDelay = 0.3
    self.TimeSinceLastCheck = 0
    self.UpdateToken = 1
    self.font = Settings.SPNewFont
    self.colourScheme = {Turbine.UI.Color(0,0,0), Turbine.UI.Color(0,0,0), Turbine.UI.Color(0,0.3,0), Turbine.UI.Color(1,1,1), Turbine.UI.Color(0,0,0), Turbine.UI.Color(0.3,0,0), Settings.TargetWindowFrontTexture, Settings.TargetWindowBackTexture, Settings.TargetWindowFrontOpacity, Settings.TargetWindowBackOpacity}

    self.PermDelay = 3

    self.activeTarget = false

    self.TimerHeight = 20

    self.ResortTimers = false

    self.targetEffects = nil
    self.target = nil
    self.targetHealth = nil
    self.targetTime = Turbine.Engine.GetGameTime()
    self.targetEffectsID = {}

    self.ShowAll = Settings.TShowAll
    self.DoEffects = Settings.TEffects
    self.pvpEffects = Settings.SKTEnabled
    self.AlertsEnabled = Settings.AEnabled
    self.pvpAlerts = Settings.SKAEnabled

    self.oldGameTime = Turbine.Engine.GetGameTime()

    self.targetEffectsDictionary = Settings.NewTargetWindowEffects

    self.callback = {}

    self.player = Turbine.Gameplay.LocalPlayer.GetInstance()

    self:SetSize(self.width, 1000)
    self:SetPosition(self.x, self.y)
    self:SetVisible(true)
    self:SetWantsUpdates(true)
    self:SetMouseVisible(false)
    self:SetOpacity(1)

    self.BackgroundWindow = Turbine.UI.Control()
    self.BackgroundWindow:SetParent(self)
    self.BackgroundWindow:SetSize(self.width , self.height)
    self.BackgroundWindow:SetBackColor(Turbine.UI.Color(0,0,0))
    self.BackgroundWindow:SetOpacity(1)
    self.BackgroundWindow:SetVisible(true)
    self.BackgroundWindow:SetMouseVisible(false)

    self.OpacityBackground = Turbine.UI.Control()
    self.OpacityBackground:SetParent(self.BackgroundWindow)
    self.OpacityBackground:SetSize(self.width - 2 * self.borderWidth, self.height - 2 * self.borderWidth)
    self.OpacityBackground:SetPosition(self.borderWidth, self.borderWidth)
    self.OpacityBackground:SetBackColor(Turbine.UI.Color(0.3, 0, 0, 0))
    self.OpacityBackground:SetVisible(true)
    self.OpacityBackground:SetMouseVisible(false)

    self.TopWindow = Turbine.UI.Window()
    self.TopWindow:SetParent(self)
    self.TopWindow:SetSize(self.width - 2 * self.borderWidth, self.height - 2 * self.borderWidth)
    self.TopWindow:SetPosition(self.borderWidth, self.borderWidth)
    self.TopWindow:SetVisible(true)
    self.TopWindow:SetMouseVisible(false)

    self.NameBackground = Turbine.UI.Control()
    self.NameBackground:SetParent(self.TopWindow)
    self.NameBackground:SetSize(self.TopWindow:GetWidth(), self.dataHeight)
    self.NameBackground:SetPosition(0, 0)
    self.NameBackground:SetBackColor(Turbine.UI.Color.Black)

    self.NameForeground = Turbine.UI.Control()
    self.NameForeground:SetParent(self.TopWindow)
    self.NameForeground:SetSize(self.TopWindow:GetWidth(), self.dataHeight)
    self.NameForeground:SetBackColor(Turbine.UI.Color(0, 0.5, 0.6))

    self.NameLabel = Turbine.UI.Label()
    self.NameLabel:SetParent(self.TopWindow)
    self.NameLabel:SetSize(self.TopWindow:GetWidth(), self.nameHeight)
    self.NameLabel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter)
    self.NameLabel:SetMouseVisible(false)
    self.NameLabel:SetMultiline(false)
    self.NameLabel:SetFont(Turbine.UI.Lotro.Font.BookAntiquaBold19)
    self.NameLabel:SetOutlineColor(Turbine.UI.Color.Black)
    self.NameLabel:SetFontStyle(Turbine.UI.FontStyle.Outline)
    --self.NameLabel:SetBackColor(Turbine.UI.Color(0.3, 0, 0, 0))

    self.DataLabel =  Turbine.UI.Label()
    self.DataLabel:SetParent(self.TopWindow)
    self.DataLabel:SetSize(self.TopWindow:GetWidth(), self.dataHeight)
    self.DataLabel:SetPosition(0, self.nameHeight)
    self.DataLabel:SetTextAlignment(Turbine.UI.ContentAlignment.MiddleCenter)
    self.DataLabel:SetMouseVisible(false)
    self.DataLabel:SetMultiline(false)
    self.DataLabel:SetFont(Turbine.UI.Lotro.Font.BookAntiquaBold18)
    self.DataLabel:SetOutlineColor(Turbine.UI.Color.Black)
    self.DataLabel:SetFontStyle(Turbine.UI.FontStyle.Outline)
    --self.DataLabel:SetBackColor(Turbine.UI.Color.White)

    self.TimerContainer = Turbine.UI.ListBox()
    self.TimerContainer:SetParent(self)
    self.TimerContainer:SetVisible(true)
    self.TimerContainer:SetSize(self.width, 1000)
    self.TimerContainer:SetPosition(0, self.height)
    self.TimerContainer:SetMouseVisible(false)

    self.SortMethod = function(a, b)
        if a:GetRemainingTime() == b:GetRemainingTime() then
            return a:GetText() < b:GetText()
        else
            return a:GetRemainingTime() > b:GetRemainingTime()
        end
    end

    local moveBox = MoveBox(self, self.width, self.height, "Target Window")
    
    self.ToggleMove = function()
        moveBox:ToggleMove()
    end

    -- Set callbacks
    self.targetCallback = TargetHandlerObject:AddTargetInfoCallback(function(args) self:TargetEvents(args) end)
    self.callback["InCombatChanged"] = AddCallback(PlayerHandlerObject, "InCombatChanged", function() self:CombatUpdate() end)

    self.healthTargetCallback = TargetHandlerObject:AddHealthCallback( 
        function(morale)
            self.newHealth = morale
            self:SetNameText()
        end)

    self.effectTargetAddedCallbacks = TargetHandlerObject:AddEffectCallback( 
        function(effect)
            self:EffectFunction(effect)
        end)

    -- Handle initial situation
    self:CombatUpdate()
    self:TargetEvents()
    self:SetNameText()
end

function TargetWindow:EffectFunction(effect)
    if not(self.targetEffectsID[effect:GetID()]) then
        if self.ShowAll or self.targetEffectsDictionary[effect:GetName()] then
            -- Only add id if not permanent effect
            --if effect:GetDuration() < 999999 then 
                
            --end
            self:NewTimer(effect)
        end
    end
end

-- Clean callbacks
function TargetWindow:RemoveCallbacks()
    TargetHandlerObject:RemoveTargetInfoCallback(self.targetCallback)
    RemoveCallback(PlayerHandlerObject, "InCombatChanged", self.callback["InCombatChanged"])
    if self.targetCallback then 
        TargetHandlerObject:RemoveTargetInfoCallback(self.targetCallback)
        self.targetCallback = nil
    end
    if self.healthTargetCallback then
        TargetHandlerObject:RemoveHealthCallback(self.healthTargetCallback)
        self.healthTargetCallback = nil
    end
    if self.effectTargetAddedCallbacks then
        TargetHandlerObject:RemoveEffectCallback(self.effectTargetAddedCallbacks)
        self.effectTargetAddedCallbacks = nil
    end
end

function TargetWindow:AdjustWindowSize(active)
    if active then
        self.BackgroundWindow:SetSize(self.width , self.height)
        self.OpacityBackground:SetSize(self.width - 2 * self.borderWidth, self.height - 2 * self.borderWidth)
    else
        self.BackgroundWindow:SetSize(self.width , self.height - self.dataHeight)
        self.OpacityBackground:SetSize(self.width - 2 * self.borderWidth, self.height - 2 * self.borderWidth - self.dataHeight)
    end
end

-- Makes or updates timer based on effect
function TargetWindow:NewTimer(effect)
    local effectName = effect:GetName()
    local numberLessName = string.match(effectName, "%D+")
    local item = self.Timers[numberLessName]

    -- Check if timer already exists
    if item then
        -- Timer already exists
        if item.permanent then
            if effectName ~= item.effectName then
                item:ChangeName(effectName)
                item.effectName = effectName
                item:ChangeIcon(effect:GetIcon())
            end
            item.permanentTimer = 0
            return
        else
            if item.ID == effect:GetID() then
                return
            else
                item.ID = effect:GetID()
                item:ChangeIcon(effect:GetIcon())
                if effectName ~= item.effectName then
                    item:ChangeName(effectName)
                    item.effectName = effectName
                end
                -- Update timer for new effect 
                item:ChangeDuration(effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime()), effect:GetDuration())
                self.ResortTimers = true
                return
            end
        end
    end

    -- Haven't seen this effect before so we make a timer
    local timer = EffectTimer(nil, self.width, self.TimerHeight, nil, effectName, self.font, self.colourScheme, (effect:GetDuration() < 999999), nil, effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime()), effect:GetIcon(), nil, nil, nil, false, 1, Turbine.UI.Color.Black)
    timer:SetTimer(effect:GetDuration() - (Turbine.Engine.GetGameTime() - effect:GetStartTime()), effect:GetDuration())
    timer.ID = effect:GetID()
    timer.effectName = effectName
    timer.permanentTimer = 0
    timer.permanent = (effect:GetDuration() > 999999)
    timer.numberLessName = numberLessName

    self.TimerContainer:AddItem(timer)
    self.Timers[numberLessName] = timer

    self.ResortTimers = true
end

-- Handles setting/resetting data of a target
function TargetWindow:TargetEvents(args)
    self:Clear()

    -- Start regathering info
    self.activeTarget = false

    if args then 

        -- Make sure we can see the window now that we have a target
        self:SetVisible(true)
        self:SetWantsUpdates(true)

        self:AdjustWindowSize(true)

        self.targetName = args.Name

        -- Check if target can have effects
        if args.Morale then

            self.activeTarget = true

            self.targetEffectsID = {}

            -- Set current data
            self.targetTime = Turbine.Engine.GetGameTime()
            self.targetHealth = args.Morale
            self.targetMaxHealth = args.MaxMorale
            self.newHealth = args.Morale

            -- Force first update
            self:CombatUpdate()
        else
            self:AdjustWindowSize(false)
        end
    else
        -- No target -> Hide window
        self:SetWantsUpdates(false)
        self:SetVisible(false)
    end

    -- Handle label of window
    self:SetNameText()
end

-- Clear all timers
function TargetWindow:Clear()
    -- Destroy all timers in the table and remove them from table
    for key, value in pairs(self.Timers) do
        self.Timers[key] = nil
        value:SetVisible(false)
        self.TimerContainer:RemoveItem(value)
    end

    -- Reset all variables
    self.targetEffects = nil
    self.targetTime = 0
    self.targetHealth = 0
    self.targetEffectsID = nil
    self.Timers = nil
    self.targetName = nil

    collectgarbage()

    self.targetEffectsID = {}
    self.Timers = {}
    self.NameForeground:SetSize(self.TopWindow:GetWidth(), self.dataHeight)
end

-- Handle an update to combat state
function TargetWindow:CombatUpdate()
    if self.player:IsInCombat() then

        -- If we enter combat and have a target setup tracking
        if self.activeTarget then
            self:SetWantsUpdates(true)
        end

        -- Get time when we start tracking combat on target
        self.targetTime = Turbine.Engine.GetGameTime()
    end
end

-- Handle setting name text with or without correct data
function TargetWindow:SetNameText()
    -- Only update if we have a target
    if self.targetName then

        -- Check if we received data from the game
        if self.activeTarget then

            -- Get target morale
            --local targetMorale = self.target:GetMorale()

            -- Get time between starting to track and now
            local delta = Turbine.Engine.GetGameTime() - self.targetTime

            -- Update if we have a health change and time since target is more than 1 second
            if self.targetHealth < self.newHealth and delta < 1 then
                self.targetHealth = self.newHealth
            end

            local percentage = math.max(self.newHealth, 1) / math.max(self.targetMaxHealth, 1)
            self.NameForeground:SetSize(self.TopWindow:GetWidth() * percentage, self.dataHeight)

            -- Format dps string
            if delta > 0 then
                self.DataLabel:SetText(string.format("%.2f%% DPS: %.2fK", percentage * 100, (self.targetHealth - self.newHealth) / delta / 1000))
            else
                self.DataLabel:SetText(string.format("%.2f%%", percentage * 100))
            end

            -- Set name
            self.NameLabel:SetText(NameHandlerObject:GetName(self.targetName, true, "TargetWindow", ""))
        else
            -- Only set name in case we haven't received data yet or will never receive
            self.NameLabel:SetText(NameHandlerObject:GetName(self.targetName, true, "TargetWindow", ""))
            self.DataLabel:SetText("")
        end
    else
        -- Set basic state
        self.NameLabel:SetText("-")
        self.DataLabel:SetText("")
    end
end

-- Update loop
function TargetWindow:Update()   
    -- What is the time?
    local currentGameTime = Turbine.Engine.GetGameTime()

    -- Get time passed since last handled update
    local delta = currentGameTime - self.oldGameTime

    -- See if we want to update the window
    if delta < self.UpdateDelay then return end

    -- Update effects
    if self.TimeSinceLastCheck > self.CheckDelay then       
        self.TimeSinceLastCheck = 0

        --Update text
        self:SetNameText()
    else
        self.TimeSinceLastCheck = self.TimeSinceLastCheck + delta
    end

    for _, timer in pairs(self.Timers) do
        -- Check for permanent timer
        if timer.permanent then
            timer.permanentTimer = timer.permanentTimer + delta
            if timer.permanentTimer > self.PermDelay then
                self.Timers[timer.numberLessName] = nil
                timer:SetVisible(false)
                self.TimerContainer:RemoveItem(timer)
            end    
        else
            -- Update timer
            timer:UpdateText(delta)
            if timer:GetRemainingTime() <= 0 then
                self.Timers[timer.numberLessName] = nil
                timer:SetVisible(false)
                self.TimerContainer:RemoveItem(timer)
            end
        end
    end

    -- Resort timers
    if self.ResortTimers then
        self.TimerContainer:Sort(self.SortMethod)
    end

    self.oldGameTime = currentGameTime
end